#include "aadl_data.h"
#include "aadl_multiarch.h"

StatusType Get_Resource(resource_t * resource)
{
  StatusType status = E_OK;
  void* access_protocol;
  switch(resource->protocol)
  {
    	case NONE:
    	  return status;
    	case LOCK:
      access_protocol = (lock_access_protocol_t*) resource->protocol_config;
    	  status = lock_data((lock_access_protocol_t*)access_protocol);
      break;
    	case ICPP:
    	  access_protocol = (icpp_access_protocol_t*) resource->protocol_config;
    	  status = icpp_lock_data((icpp_access_protocol_t*) access_protocol);
    	  break;
  }
  return status;
}

StatusType Release_Resource(resource_t * resource)
{
  StatusType status = E_OK;
  void * access_protocol;
  switch(resource->protocol)
  {
    	case NONE:
    	  return status;
    	case LOCK:

      access_protocol = (lock_access_protocol_t*) resource->protocol_config;
    	  status = unlock_data((lock_access_protocol_t*) access_protocol);
      break;
    	case ICPP:
    	  access_protocol = (icpp_access_protocol_t*) resource->protocol_config;
    	  status = icpp_unlock_data((icpp_access_protocol_t*) access_protocol);
    	  break;
  }  return status;
}

StatusType Write_Data(data_reference_t * data, void * value)
{
  StatusType status = Get_Resource(data->resource);
  if (status != E_OK)
    return status;

  memcpy(data->data, value, data->size);

  status = Release_Resource(data->resource);
  return status;
}

StatusType Read_Data(data_reference_t * data, void * dst)
{
  StatusType status = Get_Resource(data->resource);
  if (status != E_OK)
    return status;

  memcpy(dst, data->data, data->size);

  status = Release_Resource(data->resource);
  return status;
}
