#include "aadl_runtime_services.h"
#include "aadl_multiarch.h"
#include "aadl_ports.h"
#include "bitset.h"
#include "main.h"
#include "deployment.h"
#include "dimensions.h"

extern mode_id_t current_mode;

StatusType Current_System_Mode(mode_id_t * mode)
{
  StatusType ret = 0;
  // lock
  *mode = current_mode; // FIXME changes to compile
  //unlock
  return ret;
}

char Is_Thread_Active(thread_config_t * config)
{
	return (config->state != STOPPED_THREAD_STATE
			&& config->state != INITIALIZING_THREAD_STATE);
}

void Restart_Thread(thread_config_t * config)
{
	if(config->state == STOPPED_THREAD_STATE)
	{
		Start_Thread(config);
	}
}

extern char * tasks_in_mode[MODES_NB];

char Is_In_Mode(thread_config_t * config, mode_id_t mode)
{
	char * tasks = tasks_in_mode[mode];
	return BITTEST(tasks, config->id);
}

char Is_In_Current_Mode(thread_config_t * config)
{
	if(MODES_NB<=1)
		return 1;
	return Is_In_Mode(config, current_mode);
}

extern unsigned long reached_init_barrier;
//extern thread_config_addr thread_configs_in_all_modes[4];
//extern int thread_number_in_all_modes;
//extern thread_config_addr ** thread_configs_per_mode;
//extern int * thread_number_per_mode;


/**
 * Set_System_Mode should be executed in a mode transition
 * handler (thread) at higher priority than mode-dependent
 * threads.
 */
StatusType Set_System_Mode(mode_id_t * mode, mode_switch_protocol_t protocol)
{
  StatusType ret = 0;
  ret = lock_init_barrier();

  int i;
  for(i=0; i<TASKS_NB; i++) {
    thread_config_t * config = this_process.threads[i];
    unsigned long isInSrcMode = Is_In_Mode(config, current_mode);
    unsigned long isInTgtMode = Is_In_Mode(config, *mode);
#if defined RUNTIME_DEBUG && (defined USE_POK || defined USE_POSIX)
    printf("Setting mode, %s in src mode: %d; in tgt mode: %d\n", config->name, isInSrcMode, isInTgtMode);
#endif
    if(isInSrcMode && !isInTgtMode) {
      if(protocol == EMERGENCY_MODE_SWITCH) {
    	  // stop threads not in target mode
    	  Stop_Thread(config);
    	  config->state = STOPPED_THREAD_STATE;
      }
    }
  }

  current_mode = *mode;

  // count already active threads in mode
  /// and start stopped threads
  int already_active_threads=0;
  for(i=0; i<thread_number_in_all_modes;i++)
  {
	  thread_config_t * config = thread_configs_in_all_modes[i];
	  if(Is_Thread_Active(config))
		  already_active_threads++;
  }
  for(i=0; i<thread_number_per_mode[current_mode];i++)
  {
	  thread_config_t * config = thread_configs_per_mode[current_mode][i];
	  if(Is_Thread_Active(config))
	  	  already_active_threads++;
  }


  int number_of_missing_threads = thread_number_in_all_modes
		  + thread_number_per_mode[current_mode]
		  - already_active_threads;
#if defined RUNTIME_DEBUG && (defined USE_POK || defined USE_POSIX)
  printf("Wait for %d threads\n", number_of_missing_threads);
#endif

  if(number_of_missing_threads>0)
	  wait_threads_init_multiarch(number_of_missing_threads);

#if defined RUNTIME_DEBUG && (defined USE_POK || defined USE_POSIX)
  printf("All threads ready\n");
#endif

  set_start_time();

#if defined RUNTIME_DEBUG && (defined USE_POK || defined USE_POSIX)
  runtime_time_t start_t;
  get_start_time(&start_t);
  printf("Start time %u msec, %u nsec\n", start_t.msec, start_t.nsec);
#endif

  for(i=0; i<thread_number_in_all_modes;i++)
  {
	  thread_config_t * config = thread_configs_in_all_modes[i];
	  release_task_in_mode_multiarch(config);
  }
  for(i=0; i<thread_number_per_mode[current_mode];i++)
  {
  	  thread_config_t * config = thread_configs_per_mode[current_mode][i];
  	  release_task_in_mode_multiarch(config);
  }

  unlock_init_barrier();
  return ret;
}

void init_iteration_counters(thread_config_t * config, char init)
{
  switch (config->protocol)
    {
    case PERIODIC:
      {
	periodic_thread_config_t * per_thr = config->thread;
	per_thr->iteration_counter = init;
	break;
      }
    case HYBRID:
      {
	hybrid_thread_config_t * hyb_thr = config->thread;
#ifdef USE_POSIX
	hyb_thr->nb_period = init;
#endif
	break;
      }
    default:
      break;
    }
}

void set_start_time()
{
  int i;
  for(i=0; i<TASKS_NB; i++) {
    thread_config_t * config = this_process.threads[i];
    init_iteration_counters(config, 0);
  }
  set_start_time_multiarch();
}


StatusType Await_Mode(thread_config_t * config)
{
  StatusType ret = Await_Mode_multiarch(config);
  return ret;
}
