#ifndef AADL_MULTIARCH_H
#define AADL_MULTIARCH_H

#include "aadl_runtime_services.h"

#ifdef DEBUG
void print_debug(char * str);
#endif

#ifdef USE_OSEK
typedef unsigned int size_t;
extern void memcpy(void *dst, const void *src, unsigned int size);
extern size_t strlen(const char *str);
char* itoa(int num, char* str, int base);
char *strcat(char *dest, const char *src);
#endif

void get_time_multiarch(runtime_time_t * time);

StatusType release_task_in_mode_multiarch(thread_config_t * config);
StatusType create_thread_multiarch(thread_config_t * config);
StatusType start_thread_multiarch(thread_config_t * config);
StatusType stop_thread_multiarch(thread_config_t * config);
StatusType lock_init_barrier();
StatusType unlock_init_barrier();
StatusType wait_threads_init_multiarch(int threads_under_init);
StatusType set_thread_priority_multiarch(thread_config_t * config, unsigned int priority);

StatusType lock_port_reference(thread_input_port_t * port);
StatusType unlock_port_reference(thread_input_port_t * port);
StatusType send_signal(thread_input_port_t * port);

StatusType send_out_process(port_reference_t * source,
		void * value,
		unsigned int msg_size);

StatusType receive_in_process(port_reference_t * destination,
		void * received_msg);

StatusType lock_data(lock_access_protocol_t * protocol);
StatusType unlock_data(lock_access_protocol_t * protocol);

StatusType icpp_lock_data(icpp_access_protocol_t * protocol);
StatusType icpp_unlock_data(icpp_access_protocol_t * protocol);

int init_port_list_mutex(port_list_t * info);
StatusType lock_port_list(port_list_t * port_list);
StatusType unlock_port_list(port_list_t * port_list);
StatusType wait_port_list(port_list_t * port_list);
void sporadic_thread_sleep(sporadic_thread_config_t * info);
void set_start_time_multiarch();

int init_thread_config_multiarch(thread_config_t * config);

int init_periodic_config_multiarch(periodic_thread_config_t * info);
StatusType Await_Mode_multiarch(thread_config_t * config);
StatusType await_periodic_dispatch_multiarch(periodic_thread_config_t * info);

StatusType init_hybrid_timer(hybrid_thread_config_t * info);
StatusType hybrid_reset_timer(hybrid_thread_config_t * info);
StatusType timedwait_hybrid(hybrid_thread_config_t * info);

StatusType timedwait_timed_thread(timed_thread_config_t * info);

StatusType init_time_triggered_config(time_triggered_thread_config_t * config);
StatusType await_time_triggered_dispatch_multiarch(time_triggered_thread_config_t * info);

#endif // AADL_MULTIARCH_H
