#include "aadl_ports.h"

extern process_config_t this_process;

StatusType Send_Output_Destination(port_connection_t * connection) {
	StatusType status = E_OK;
	port_reference_t * destination = connection->destination;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
	printf("Send output destination type = %d\n", connection->destination->parent_type);
#endif
	if(destination->parent_type == THREAD) {
	  	thread_input_port_t * in_port = (thread_input_port_t *) connection->destination->directed_port;
		in_port->fresh=1;
		status = Send_Output_Inter_Thread(connection);
		if(status!=E_OK) {
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Send_Output_Destination: ERROR, local send_output failed\n");
#endif
			return status;
		}
	}
	else if(destination->parent_type == PROCESS) {
		status = Send_Output_Inter_Process(connection);
		if(status!=E_OK) {
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Send_Output: ERROR, inter process send_output failed\n");
#endif
			return status;
		}
	}
	return status;
}

StatusType Send_Output_Inter_Process(port_connection_t * connection)
{
  port_reference_t * source = connection->source;
  thread_output_port_t * out_port = (thread_output_port_t*) source->directed_port;
  void * value = out_port->port_variable;
  unsigned int msg_size = source->msg_size;
  return send_out_process(connection->destination, value, msg_size);
}

StatusType Send_Output_Inter_Thread_Sampled(port_connection_t * connection)
{
	StatusType status = E_OK;

	thread_input_port_t * in_port = (thread_input_port_t *) connection->destination->directed_port;
	status = lock_port_reference(in_port);
	if (status != E_OK)
	{
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Send_Output_Inter_Thread: ERROR %d when acquiring lock\n", status);
#endif
		return status;
	}

	status = Send_Output_Inter_Thread_Fifo(connection);

	status = unlock_port_reference(in_port);
	if (status != E_OK)
		Error_Handler(status, in_port->parent->parent, NULL);

	return status;
}


StatusType Send_Output_Inter_Thread_PDP_LowMFP(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * source = connection->source;
	port_reference_t * destination = connection->destination;

	thread_input_port_t * in_port = (thread_input_port_t *) destination->directed_port;

	PDP_lowMFP_attr_t * attr = (PDP_lowMFP_attr_t *) connection->protocol->protocol_attr;

	thread_output_port_t * out_port = (thread_output_port_t*) source->directed_port;
	void * value = out_port->port_variable;

	thread_config_t * thread_config = (thread_config_t *) source->parent;
	periodic_thread_config_t * periodic_config = thread_config->thread;
	int iteration_counter = periodic_config->iteration_counter % attr->write_lookup_table_size;

	int tmp;
	int SEJD = 0;
	int CDW = 0;

	unsigned int writer_task_period = periodic_config->period;
	unsigned int writer_task_deadline = thread_config->deadline;
	unsigned int writing_time = iteration_counter * writer_task_period + writer_task_deadline;
	int simultaneous_writers = 0;
	int other_period, other_deadline;
	int cnx_idx;
	char itself = 0;
	thread_config_t * other_writer;
	for(cnx_idx=0; cnx_idx<in_port->connected_ports_length;cnx_idx++)
	{
	    other_writer = in_port->sources[cnx_idx]->source->parent;
	    periodic_thread_config_t * other_periodic_writer = (periodic_thread_config_t*) other_writer->thread;
	    other_period = other_periodic_writer->period;
	    other_deadline = other_writer->deadline;
	    // (t - Dj)/Pj can be negative, and its floor equals to -1
	    // we don't handle cases where it equals less than -1...
	    if(other_writer->id==thread_config->id)
	      itself = 1;
	    if ((tmp = writing_time - other_deadline) >= 0) {
	      SEJD += (tmp / other_period)+1; // + 1 ? 
	    }
	    else
	    {
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
	      printf("ERROR, negative writing time %d \n",writing_time - other_deadline );
#endif
	    }
	    if (((writing_time - other_deadline) % other_period) == 0
	    		&& itself==1) {
	      simultaneous_writers++;
	    }
	}
	// TODO add mode specific writers

	CDW = SEJD - simultaneous_writers - 1; // - 1 to write at the beginning of the queue
	// TODO data port
	if (destination->type == EVENT_DATA) {
		event_data_port_t * edp = in_port->content;
		char * buf = (char * ) edp->offset;
		CDW = CDW % edp->queue_size;
		if(CDW < 0)
			CDW += edp->queue_size;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("send in %d\n", CDW);
#endif
		memcpy(&buf[CDW*(destination->msg_size/sizeof(char))],
				value,
				destination->msg_size);
	}
	// TODO event port
	return status;
}

StatusType Send_Output_Inter_Thread_PDP_LowET(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * source = connection->source;
	port_reference_t * destination = connection->destination;

	thread_input_port_t * in_port = (thread_input_port_t *) destination->directed_port;

	PDP_lowET_attr_t * attr = (PDP_lowET_attr_t *) connection->protocol->protocol_attr;

	thread_output_port_t * out_port = (thread_output_port_t*) source->directed_port;
	void * value = out_port->port_variable;

	thread_config_t * thread_config = (thread_config_t *) source->parent;
	periodic_thread_config_t * periodic_config = thread_config->thread;

	if (destination->type == DATA) {
		data_port_t * port_data = in_port->content;

		int iteration_counter = periodic_config->iteration_counter%attr->write_lookup_table_size;
		mode_id_t current_mode;
		Current_System_Mode(&current_mode);
		void * addr;
		if(attr->write_lookup_table[iteration_counter])
			addr = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].writing_data;
		else
			addr = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].reading_data;

		memcpy(addr,
				value,
				destination->msg_size);
	}
	else if (destination->type == EVENT_DATA) {
		event_data_port_t * edp = in_port->content;
		char * buf = (char * ) edp->offset;
		int iteration_counter = (periodic_config->iteration_counter)%(attr->write_lookup_table_size);
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("send in %d\n", attr->write_lookup_table[iteration_counter]);
#endif
		memcpy(&buf[attr->write_lookup_table[iteration_counter]*(destination->msg_size/sizeof(char))],
							value,
							destination->msg_size);
	}
	else if (destination->type == EVENT) {
		event_port_t * ep = in_port->content;
		return -1; // TODO: copy edp implementation when tested
	}
    return status;
}

char Is_Full(port_reference_t * port)
{
  int nb_msg=0;
  thread_input_port_t * in_port = (thread_input_port_t*) port->directed_port;
  if (port->type == EVENT_DATA) {
    event_data_port_t * port_ed = in_port->content;
    if(port_ed->beg_read_idx > port_ed->end_read_idx)
      nb_msg += (port_ed->queue_size - port_ed->beg_read_idx + port_ed->end_read_idx);
    else
      nb_msg += (port_ed->end_read_idx - port_ed->beg_read_idx);
    if(port_ed->end_write_idx < port_ed->end_read_idx)
      nb_msg += (port_ed->queue_size - port_ed->end_read_idx + port_ed->end_write_idx);
    else
      nb_msg += (port_ed->end_write_idx - port_ed->end_read_idx);
    return (nb_msg >= port_ed->queue_size);
  }
  else if (port->type == EVENT) {
    event_port_t * port_ed = in_port->content;
    if(port_ed->beg_read_idx > port_ed->end_read_idx)
      nb_msg += (port_ed->queue_size - port_ed->beg_read_idx + port_ed->end_read_idx);
    else
      nb_msg += (port_ed->end_read_idx - port_ed->beg_read_idx);
    if(port_ed->end_write_idx < port_ed->end_read_idx)
      nb_msg += (port_ed->queue_size - port_ed->end_read_idx + port_ed->end_write_idx);
    else
      nb_msg += (port_ed->end_write_idx - port_ed->end_read_idx);
    return (nb_msg >= port_ed->queue_size);
  }
  return 0;
}

StatusType Send_Output_Inter_Thread_Fifo(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * source = connection->source;
	port_reference_t * destination = connection->destination;

	thread_input_port_t * in_port = (thread_input_port_t *) destination->directed_port;
	thread_output_port_t * out_port = (thread_output_port_t*) source->directed_port;
	void * value = out_port->port_variable;

	if (destination->type == DATA)
	{
		// FIXME change here to include the size of the data ?
		data_port_t * port_data = in_port->content;
		mode_id_t current_mode;
		status = Current_System_Mode(&current_mode);
		void * writing_addr = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].writing_data;
		memcpy(writing_addr,
				value,
				destination->msg_size);
		in_port->parent->msg_nb++;

#if defined RUNTIME_DEBUG && defined USE_POSIX
		printf("Write in index %d, %d (dataport)\n", writing_addr, *((char*) value));
#endif
	}
	else if (destination->type == EVENT_DATA)
	{
		event_data_port_t * port_ed = in_port->content;
		if (Is_Full(destination)) {
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Send_Output_Inter_Thread: ERROR, full queue\n");
#endif
			return E_OS_LIMIT;
		} else {

			in_port->parent->msg_nb++;

			if(in_port->parent->waiting)
				send_signal(in_port);

			char * buf = (char * ) port_ed->offset;
			memcpy(&buf[port_ed->end_write_idx*(destination->msg_size/sizeof(char))],
					value,
					destination->msg_size);
			port_ed->end_write_idx = (port_ed->end_write_idx+1) % port_ed->queue_size;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Write in index %d, (event data port)\n", port_ed->end_write_idx/*, *((char*) value)*/);
#endif
		}
	}
	else
	{
		event_port_t * port_event = in_port->content;
		if (Is_Full(destination)) {
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
		  printf("Send_Output_Inter_Thread: ERROR, full queue\n");
#endif
			return E_OS_LIMIT;
		} else {
			in_port->parent->msg_nb++;
			if(in_port->parent->waiting)
				send_signal(in_port);

			port_event->end_write_idx = (port_event->end_write_idx+1) % port_event->queue_size;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Write in index %d, %d \n", port_event->end_write_idx, *((char*) value));
#endif
		}
	}
	return status;
}

StatusType Send_Output_Inter_Thread(port_connection_t * connection)
{
  switch(connection->protocol->protocol_type) {
  	  case SAMPLED :
  		  return Send_Output_Inter_Thread_Sampled(connection);
  	  case IMMEDIATE :
  		  return Send_Output_Inter_Thread_Fifo(connection);
  	  case PDP_LOWET:
  		  return Send_Output_Inter_Thread_PDP_LowET(connection);
  	  case PDP_LOWMFP:
  		  return Send_Output_Inter_Thread_PDP_LowMFP(connection);
  	  default:
  		return -1;
  }
}

StatusType Do_Send_Output_Data_Thread(thread_config_t * thread)
{
	StatusType status = E_OK;
	int i;
	for (i = 0; i < thread->out_ports_size; i++)
	{
	  port_reference_t * src_port = thread->out_ports[i];
	  if(src_port->type != DATA)
		  continue;
	  status = Do_Send_Output(src_port);
	  if (status != E_OK)
	    return status;
	}
	return status;
}

StatusType Do_Send_Output(port_reference_t * src_port)
{
	StatusType status = E_OK;
	thread_output_port_t * out_port = (thread_output_port_t*) src_port->directed_port;

	int i;
	for(i=0; i<out_port->connected_ports_length; i++) {
		port_connection_t * connection = out_port->destinations[i];
		status = Send_Output_Destination(connection);
		if(status!=E_OK) {
			return status;
		}
	}
	mode_id_t current_mode;
	status = Current_System_Mode(&current_mode);

	int j;
	for (j = out_port->dest_index_by_mode[current_mode];
			j < out_port->dest_index_by_mode[current_mode+1];
			j++)
	{
		port_connection_t * connection = out_port->destinations[j];
		status = Send_Output_Destination(connection);
		if (status != E_OK)
			return status;
	}
	return status;
}

StatusType Send_Output_Thread(thread_config_t * thread)
{
  StatusType status = E_OK;
  int i;
  for (i = 0; i < thread->out_ports_size; i++)
  {
    status = Send_Output((thread->out_ports)[i]);
    if (status != E_OK)
      return status;
  }
  return status;
}

StatusType Receive_Input_Thread(thread_config_t * thread)
{
  StatusType status = E_OK;
  int i;
  for (i = 0; i < thread->in_ports_size; i++)
  {
	status = Receive_Input((thread->in_ports)[i]);
	if (status != E_OK)
		return status;
  }
  return status;
}

StatusType Receive_Input_Inter_Thread_Fifo(port_connection_t * connection)
{
	StatusType status = E_OK;
	int discarded_msg_nb = 0;

	port_reference_t * destination = connection->destination;
    port_reference_t * source = connection->source;

	thread_input_port_t * in_port = (thread_input_port_t*) destination->directed_port;

	if (destination->type == DATA)
	{
		data_port_t * port_data = in_port->content;
		if (in_port->fresh)
		{
			mode_id_t current_mode;
			Current_System_Mode(&current_mode);
			void * read_addr = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].reading_data;
			port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].reading_data = port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].writing_data;
			port_data->data_rw_per_mode[port_data->src_index_by_mode[current_mode]].writing_data = read_addr;
			discarded_msg_nb++;
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("Receive input local on data port\n");
#endif

		}
	}
	else if (destination->type == EVENT_DATA)
	{
		event_data_port_t * port_ed = in_port->content;
		if(port_ed->end_read_idx < port_ed->beg_read_idx)
			discarded_msg_nb = port_ed->queue_size-port_ed->beg_read_idx+port_ed->end_read_idx;
		else
			discarded_msg_nb = port_ed->end_read_idx-port_ed->beg_read_idx;

		port_ed->beg_read_idx = port_ed->end_read_idx;
		port_ed->end_read_idx = port_ed->end_write_idx;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Receive inputs local on event data port; first: %d, last: %d\n", port_ed->beg_read_idx, port_ed->end_read_idx);
#endif
	}
	else
	{
		event_port_t * port_event = in_port->content;
		if(port_event->end_read_idx < port_event->beg_read_idx)
			discarded_msg_nb = port_event->queue_size-port_event->beg_read_idx+port_event->end_read_idx;
		else
			discarded_msg_nb = port_event->end_read_idx-port_event->beg_read_idx;

		port_event->beg_read_idx = port_event->end_read_idx;
		port_event->end_read_idx = port_event->end_write_idx;

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Receive inputs local on event port; first: %d, last: %d\n", port_event->beg_read_idx, port_event->end_read_idx);
#endif
	}

#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
	printf("Receive inputs local: number of discarded messages is %d\n", discarded_msg_nb);
#endif

	in_port->parent->msg_nb-=discarded_msg_nb;

	return status;
}

StatusType Receive_Input_Inter_Thread_Sampled(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * destination = connection->destination;

	thread_input_port_t * in_port = (thread_input_port_t*) destination->directed_port;
	status = lock_port_reference(in_port);
	if (status != E_OK)
	{
		Error_Handler(status, in_port->parent->parent, NULL);
#if defined DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Receive inputs local: ERROR %d when acquiring lock\n", status);
#endif
		return status;
	}

	status = Receive_Input_Inter_Thread_Fifo(connection);

	status = unlock_port_reference(in_port);
	if (status != E_OK)
	{
#if defined DEBUG && defined USE_POSIX
		printf("Receive inputs local: ERROR %d when releasing lock\n", status);
#endif
	}
	return status;
}



StatusType Receive_Input_Inter_Thread_PDP_LowMFP(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * destination = connection->destination;
	port_reference_t * source = connection->source;

	thread_input_port_t * in_port = (thread_input_port_t*) destination->directed_port;

	thread_config_t * thread_config = (thread_config_t *) destination->parent;
	periodic_thread_config_t * periodic_config = thread_config->thread;

	PDP_lowMFP_attr_t * attr = (PDP_lowMFP_attr_t *) connection->protocol->protocol_attr;

	int CPR = 0, NPR=0;
	int reader_task_period = periodic_config->period;
	int iteration_counter = periodic_config->iteration_counter % attr->read_lookup_table_size;

	int reading_time = iteration_counter * reader_task_period;
	int next_reading_time = ((periodic_config->iteration_counter+1)%attr->read_lookup_table_size) * reader_task_period;

	int writer_period, writer_deadline;
	int idx;
	thread_config_t * writer;
	// TODO miss modes for input port connection references (writers might be mode specific)
	for (idx=0 ; idx < in_port->connected_ports_length ; idx++) {
		writer = in_port->sources[idx]->source->parent;
		periodic_thread_config_t * periodic_writer = (periodic_thread_config_t*) writer->thread;
		writer_period = periodic_writer->period;
		writer_deadline = writer->deadline;
		CPR += ((this_process.hyperperiod+reading_time - writer_deadline) / writer_period) + 1;
		NPR += ((this_process.hyperperiod+next_reading_time - writer_deadline) / writer_period) + 1;
	}

	static char after_hyperperiod=0;

	if(!after_hyperperiod && thread_config->protocol == PERIODIC) {
		periodic_thread_config_t * periodic_config = (periodic_thread_config_t *) thread_config->thread;
		if(periodic_config->iteration_counter*periodic_config->period>=this_process.hyperperiod)
			after_hyperperiod = 1;
	}

	if (source->type == EVENT_DATA) {
		event_data_port_t * port_ed = in_port->content;
		CPR = CPR % port_ed->queue_size;
		NPR = NPR % port_ed->queue_size;
		if(CPR < 0)
			CPR += port_ed->queue_size;
		if(NPR < 0)
			NPR += port_ed->queue_size;

		char fresh = (NPR!=CPR);

		if(fresh && after_hyperperiod) {
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("CPR = %d\n", CPR);
#endif
			port_ed->beg_read_idx = CPR; // -1 because get_value takes beg_read_idx+1;
			port_ed->end_read_idx = NPR;
		}
		else {
			port_ed->end_read_idx = port_ed->beg_read_idx;
			return E_OS_LIMIT;
		}
	} else {
		return -1; // implement when tested on event data ports
	}

	return status;
}

StatusType Receive_Input_Inter_Thread_PDP_LowET(port_connection_t * connection)
{
	StatusType ret = E_OK;

	port_reference_t * destination = connection->destination;
	port_reference_t * source = connection->source;

	thread_input_port_t * in_port = (thread_input_port_t*) destination->directed_port;

	thread_config_t * thread_config = (thread_config_t *) destination->parent;
	periodic_thread_config_t * periodic_config = thread_config->thread;
	PDP_lowET_attr_t * protocol_attr = (PDP_lowET_attr_t *) connection->protocol->protocol_attr;
	unsigned int iteration_counter = periodic_config->iteration_counter%protocol_attr->read_lookup_table_size;
	unsigned int next = (iteration_counter + 1) % protocol_attr->read_lookup_table_size;
	char fresh = (protocol_attr->read_lookup_table[next] != protocol_attr->read_lookup_table[iteration_counter]);
	// nothing to do in case of data port
	static char after_hyperperiod=0;
	if(!after_hyperperiod && thread_config->protocol == PERIODIC) {
		periodic_thread_config_t * periodic_config = (periodic_thread_config_t *) thread_config->thread;
		if(periodic_config->iteration_counter*periodic_config->period>=this_process.hyperperiod)
			after_hyperperiod = 1;
	}


	if (source->type == EVENT_DATA) {
		event_data_port_t * port_ed = in_port->content;
		if(fresh && after_hyperperiod) {
			port_ed->beg_read_idx = protocol_attr->read_lookup_table[iteration_counter];
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
			printf("CPR = %d\n", port_ed->beg_read_idx+1);
#endif
			//-1 at the end not useful if not present above
			port_ed->end_read_idx = protocol_attr->read_lookup_table[next];
		}
		else {
			port_ed->end_read_idx = port_ed->beg_read_idx;
			return E_OS_LIMIT;
		}
	} else {
		return -1; // implement when tested on event data ports
	}
	return ret;
}

StatusType Receive_Input_Inter_Thread(port_connection_t * connection)
{
	StatusType status = E_OK;
	switch(connection->protocol->protocol_type) {
		case SAMPLED :
			return Receive_Input_Inter_Thread_Sampled(connection);
		case IMMEDIATE :
			return Receive_Input_Inter_Thread_Fifo(connection);
		case PDP_LOWET:
			return Receive_Input_Inter_Thread_PDP_LowET(connection);
		case PDP_LOWMFP:
			return Receive_Input_Inter_Thread_PDP_LowMFP(connection);
		default:
			return -1;
	}

	return status;
}

StatusType Receive_Input_Inter_Process(port_connection_t * connection)
{
	StatusType status=E_OK;
	port_reference_t * source = connection->source;
	port_reference_t * destination = connection->destination;

	thread_input_port_t * in_port = (thread_input_port_t*) destination->directed_port;

	void * dest_addr = NULL;
	if(source->type==DATA)
	{
		data_port_t * dp = (data_port_t *) in_port->content;
		mode_id_t current_mode;
		Current_System_Mode(&current_mode);

		dest_addr = dp->data_rw_per_mode[dp->src_index_by_mode[current_mode]].reading_data;
		status = receive_in_process(source,
				dest_addr);
		if(status==E_OK)
		{
		  in_port->fresh=1;
		  in_port->parent->msg_nb++;
		}
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Receive inputs remote on data port; %d messages\n", in_port->parent->msg_nb);
#endif
	}
	else if(source->type==EVENT_DATA)
	{
		event_data_port_t * edp = (event_data_port_t*) in_port->content;
		char * buf = (char *) edp->offset;
		edp->beg_read_idx = edp->end_read_idx;

		while(status==E_OK)
		{
			unsigned int idx = (edp->end_read_idx)%edp->queue_size;
			dest_addr = &buf[idx*source->msg_size];
			status = receive_in_process(source,
					dest_addr);
			if(status==E_OK)
			{
				edp->end_read_idx = (edp->end_read_idx+1) % edp->queue_size;
				in_port->parent->msg_nb++;
			}
		}
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
		printf("Receive inputs remote on event data port; %d messages\n", in_port->parent->msg_nb);
#endif

	}
	else
	{
		event_port_t * ep = (event_port_t*) in_port->content;
		char * buf = (char *) ep->offset;
		ep->beg_read_idx = ep->end_read_idx;

		while(status==E_OK)
		{
			unsigned int idx = (ep->end_read_idx)%ep->queue_size;
			dest_addr = &buf[idx*source->msg_size];
			status = receive_in_process(source,
					dest_addr);
			if(status==E_OK)
			{
				ep->end_read_idx= (ep->end_read_idx+1) % ep->queue_size;
				in_port->parent->msg_nb++;
			}
		}
	}

	return status;
}

StatusType Receive_Input_Source(port_connection_t * connection)
{
	StatusType status = E_OK;
	port_reference_t * source = connection->source;
	
#if defined RUNTIME_DEBUG && (defined USE_POSIX || defined USE_POK)
	printf("Receive_Input_Source type = %d\n", connection->source->parent_type);
#endif
	if(source->parent_type == THREAD) {
		status = Receive_Input_Inter_Thread(connection);
	}
	else if(source->parent_type == PROCESS) {
		status = Receive_Input_Inter_Process(connection);
	}
	return status;

}
