#include "aadl_multiarch.h"
#include "arinc653/blackboard.h"

#include "main.h"

void __stack_chk_fail(){}


int init_thread_config_multiarch(thread_config_t * config)
{
  return 0; // init wait_mode_rez and wait_mode_event
}

int init_periodic_config_multiarch(periodic_thread_config_t * info)
{
  return 0; // nothing to do
}

void get_time_multiarch(runtime_time_t * time)
{
	StatusType ret = E_OK;
	SYSTEM_TIME_TYPE res=0;
	unsigned long long current_time = 0;
	GET_TIME(&res, &ret);
	current_time = (unsigned long long) res;
	time->msec = current_time/1000000;
	time->nsec = current_time-(time->msec*1000000);
}

runtime_time_t start_t;

void get_start_time(runtime_time_t * time)
{
	time->msec = start_t.msec;
	time->nsec = start_t.nsec;
	return;
}

extern SEMAPHORE_ID_TYPE init_barrier_event;
extern SEMAPHORE_ID_TYPE init_barrier_mutex;

extern unsigned long reached_init_barrier;
extern mode_id_t current_mode;

StatusType Await_Mode_multiarch(thread_config_t * config)
{
  StatusType ret = E_OK;
  if(config->state==INITIALIZING_THREAD_STATE
  		  && Is_In_Mode(config, current_mode))
  {
	  reached_init_barrier++;
#if defined RUNTIME_DEBUG
	  printf("Thread %s reached barrier, total=%d threads arrived\n", config->name, reached_init_barrier);
#endif
  }

  config->state=AWAITING_MODE_THREAD_STATE;

  SIGNAL_SEMAPHORE(init_barrier_event,
		    &ret);
  SIGNAL_SEMAPHORE(init_barrier_mutex,
  		    &ret);

#if defined RUNTIME_DEBUG
	  printf("Thread %s about to wait mode\n", config->name, reached_init_barrier);
#endif
  WAIT_SEMAPHORE (*(config->wait_mode_event),
		      INFINITE_TIME_VALUE,
  		      &ret);
  return ret;
}

StatusType await_periodic_dispatch_multiarch(periodic_thread_config_t * info)
{
  StatusType ret;
  info->iteration_counter++;
  PERIODIC_WAIT(&ret);
  return ret;
}

void sporadic_thread_sleep(sporadic_thread_config_t * info)
{
  int ret=0;
  // relative timeout
  unsigned long elapsed_time = info->parent->finish_time.msec - info->parent->dispatch_time.msec;
  if(elapsed_time > info->time_interval)
	  return;
#ifdef RUNTIME_DEBUG
  printf("Sporadic task %s about to wait for %d microseconds", info->parent->name, (info->time_interval - elapsed_time)*1000);
#endif
  TIMED_WAIT((SYSTEM_TIME_TYPE)((info->time_interval - elapsed_time)*1000), &ret);
}

StatusType await_time_triggered_dispatch_multiarch(time_triggered_thread_config_t * config)
{
  StatusType ret = 0;
  WAIT_SEMAPHORE (*(config->release_event),
  		      INFINITE_TIME_VALUE,
    		  &ret);
  return ret;
}

StatusType init_time_triggered_config(time_triggered_thread_config_t * config)
{
  // Nothing to do here
  StatusType ret=0;
  return ret;
}

StatusType stop_thread_multiarch(thread_config_t * config)
{
  StatusType ret=0;
#ifdef RUNTIME_DEBUG
  printf("About to stop thread %s\n", config->name);
#endif
  PROCESS_ID_TYPE * tid = (PROCESS_ID_TYPE*) config->thread_id;
  STOP(*tid, &ret);
  return ret;
}

StatusType init_hybrid_timer(hybrid_thread_config_t * info)
{
  return 0;
}

StatusType hybrid_reset_timer(hybrid_thread_config_t * info)
{
  return 0;
}

StatusType timedwait_hybrid(hybrid_thread_config_t * info)
{
  StatusType ret = 0;
  return ret;
}

StatusType timedwait_timed_thread(timed_thread_config_t * info)
{
  StatusType ret = 0;
#if defined(USE_POK) && defined(POK_NEEDS_ARINC653_SEMAPHORE)
  int unlock_ret = unlock_port_list(info->parent->global_q);
#ifdef RUNTIME_DEBUG
  printf("Timed thread wait %d\n", info->time_interval);
#endif
  WAIT_SEMAPHORE (*(info->parent->global_q->event),
          info->time_interval,
		  &ret);
#ifdef RUNTIME_DEBUG
  printf("Timed thread released\n");
#endif
  if(!unlock_ret)
	  lock_port_list(info->parent->global_q);
#endif
  return ret;
}



void set_start_time_multiarch()
{
  SYSTEM_TIME_TYPE t;
  RETURN_CODE_TYPE ret;
  GET_TIME(&t, &ret);
  unsigned long long current_time = (unsigned long long) t;
  start_t.msec = current_time/1000000;
  start_t.nsec = current_time-(start_t.msec*1000000);
}

StatusType lock_init_barrier()
{
	StatusType ret;
	WAIT_SEMAPHORE(init_barrier_mutex,
					  INFINITE_TIME_VALUE,
					  &ret);
	return ret;
}

StatusType unlock_init_barrier()
{
	StatusType ret;
	SIGNAL_SEMAPHORE(init_barrier_mutex,
					  &ret);
	return ret;
}

StatusType wait_threads_init_multiarch(int threads_under_init)
{
	StatusType ret = E_OK;
	while(reached_init_barrier<threads_under_init)
	{
		SIGNAL_SEMAPHORE(init_barrier_mutex,
				  &ret);
#ifdef RUNTIME_DEBUG
		printf("In mode manager thread, wait %d threads init, %d arrived\n", threads_under_init, reached_init_barrier);
#endif
		WAIT_SEMAPHORE (init_barrier_event,
			      INFINITE_TIME_VALUE,
	  		      &ret);
#ifdef RUNTIME_DEBUG
		printf("Mode manager woke up, passed barrier\n");
#endif
		WAIT_SEMAPHORE(init_barrier_mutex,
							  INFINITE_TIME_VALUE,
							  &ret);
	}
	return ret;
}

StatusType release_task_in_mode_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;
#if defined RUNTIME_DEBUG
	printf("Thread %s released in current mode\n", config->name);
#endif
	SIGNAL_SEMAPHORE(*(config->wait_mode_event), &ret);
	return ret;
}

StatusType create_thread_multiarch(thread_config_t * config)
{
	StatusType ret = E_OK;

	PROCESS_ATTRIBUTE_TYPE tattr;
	tattr.ENTRY_POINT = config->compute_entrypoint;

	if(config->protocol==PERIODIC)
	{
		periodic_thread_config_t * ptc =
				(periodic_thread_config_t*) config->thread;
		tattr.PERIOD = ptc->period;

		// TODO: retrieve deadline
		tattr.DEADLINE = ptc->period;

		// TODO: retreive WCET
		tattr.TIME_CAPACITY = ptc->period;
	}
	else
	{
		tattr.PERIOD = INFINITE_TIME_VALUE;

		sporadic_thread_config_t * stc =
						(sporadic_thread_config_t*) config->thread;

		tattr.DEADLINE = INFINITE_TIME_VALUE;

		// TODO: retreive WCET
		tattr.TIME_CAPACITY = INFINITE_TIME_VALUE;
	}
	tattr.BASE_PRIORITY = config->priority;
	strcpy(tattr.NAME, config->name);

	tattr.STACK_SIZE = 100000;
	CREATE_PROCESS (&(tattr), (PROCESS_ID_TYPE*) config->thread_id, &(ret));

	return ret;
}
