#include "ecrobot_interface.h"

#include "aadl_multiarch.h"

void memcpy(void *dst, const void *src, unsigned int size)
{
  char *p = dst + size;
  while (dst != p) {
    *((char*)dst) = *((char*)src);
    dst++;
    src++;
  }
}

size_t strlen(const char *str)
{
  unsigned int i=0;
  while(str[i]!='\0')
	i++;
  return i-1;
}

char * strcat(char *dest, const char *src)
{
    size_t i,j;
    for (i = 0; dest[i] != '\0'; i++)
        ;
    for (j = 0; src[j] != '\0'; j++)
        dest[i+j] = src[j];
    dest[i+j] = '\0';
    return dest;
}

void swap(char c1, char c2)
{
  char temp = c1;
  c1 = c2;
  c2 = temp;
}

/* A utility function to reverse a string  */
void reverse(char str[], int length)
{
    int start = 0;
    int end = length -1;
    while (start < end)
    {
        swap(*(str+start), *(str+end));
        start++;
        end--;
    }
}

// Implementation of itoa()
char* itoa(int num, char* str, int base)
{
    int i = 0;
    char isNegative = 0;

    /* Handle 0 explicitely, otherwise empty string is printed for 0 */
    if (num == 0)
    {
        str[i++] = '0';
        str[i] = '\0';
        return str;
    }

    // In standard itoa(), negative numbers are handled only with
    // base 10. Otherwise numbers are considered unsigned.
    if (num < 0 && base == 10)
    {
        isNegative = 1;
        num = -num;
    }

    // Process individual digits
    while (num != 0)
    {
        int rem = num % base;
        str[i++] = (rem > 9)? (rem-10) + 'a' : rem + '0';
        num = num/base;
    }

    // If number is negative, append '-'
    if (isNegative)
        str[i++] = '-';

    str[i] = '\0'; // Append string terminator

    // Reverse the string
    reverse(str, i);

    return str;
}

#ifdef DEBUG
void print_debug(char* str)
{
  return;
}
#endif

StatusType lock_port_reference(thread_input_port_t * port)
{
  return GetResource(*(port->parent->rez));
}

StatusType unlock_port_reference(thread_input_port_t * port)
{
  return ReleaseResource(*(port->parent->rez));
}

StatusType send_signal(thread_input_port_t * port)
{
  TaskType * tid = (TaskType *) port->parent->parent->thread_id;
  return SetEvent(*tid, *(port->parent->event));
}

StatusType send_out_process(port_reference_t * source,
		void * value,
		unsigned int msg_size)
{
	return INVALID_SERVICE_CALL; // Not available on osek.
}

StatusType receive_in_process(port_reference_t * destination,
		void * received_msg)
{
	return INVALID_SERVICE_CALL; // Not available on osek.
}

StatusType lock_data(lock_access_protocol_t * protocol)
{
  return GetResource(*(protocol->rez));
}

StatusType unlock_data(lock_access_protocol_t * protocol)
{
  return ReleaseResource(*(protocol->rez));
}

StatusType icpp_lock_data(icpp_access_protocol_t * protocol)
{
  return -1; // ICPP is not supported by OSEK platforms
}

StatusType icpp_unlock_data(icpp_access_protocol_t * protocol)
{
  return -1; // ICPP is not supported by OSEK platforms
}

int init_port_list_mutex(port_list_t * info)
{
  return 0;
}

StatusType lock_port_list(port_list_t * port_list)
{
  return GetResource(*(port_list->rez));
}

StatusType unlock_port_list(port_list_t * port_list)
{
  return ReleaseResource(*(port_list->rez));
}

StatusType wait_port_list(port_list_t * port_list)
{
  ReleaseResource(*(port_list->rez));
  return WaitEvent(*(port_list->event));
}

StatusType await_time_triggered_dispatch_multiarch(time_triggered_thread_config_t * config)
{
	// not implemented yet
	return -100;
}

StatusType init_time_triggered_config(time_triggered_thread_config_t * config)
{
	// nothing to do here
	return 0;
}
