#include <pthread.h>
#include <time.h>
#include <signal.h>
#include "aadl_multiarch.h"

#ifdef DEBUG
#include <stdio.h>

void print_debug(char * str)
{
  printf("%s", str);
}
#endif

extern struct timespec start_timespec;

// for aadl_port
StatusType lock_port_reference(thread_input_port_t * port)
{
#if defined RUNTIME_DEBUG && defined USE_POSIX
	printf("Lock port reference\n");
#endif
  return pthread_mutex_lock(port->parent->rez);
}

StatusType unlock_port_reference(thread_input_port_t * port)
{
#if defined RUNTIME_DEBUG && defined USE_POSIX
	printf("Unlock port reference\n");
#endif
  return pthread_mutex_unlock(port->parent->rez);
}

StatusType send_signal(thread_input_port_t * port)
{
  return pthread_cond_signal(port->parent->event);
}

StatusType send_out_process(port_reference_t * source,
		void * value,
		unsigned int msg_size)
{
	return INVALID_SERVICE_CALL; // Not available on osek.
}

StatusType receive_in_process(port_reference_t * destination,
		void * received_msg)
{
	return INVALID_SERVICE_CALL; // Not available on osek.
}

StatusType lock_data(lock_access_protocol_t * protocol)
{
  return pthread_mutex_lock(protocol->rez);
}

StatusType unlock_data(lock_access_protocol_t * protocol)
{
  return pthread_mutex_unlock(protocol->rez);
}

StatusType icpp_lock_data(icpp_access_protocol_t * protocol)
{
  return lock_data(protocol->lock);
}

StatusType icpp_unlock_data(icpp_access_protocol_t * protocol)
{
  return unlock_data(protocol->lock);
}

int init_port_list_mutex(port_list_t * info)
{
  int ret = pthread_cond_init(info->event, NULL);
  return pthread_mutex_init(info->rez, NULL);
}

StatusType lock_port_list(port_list_t * port_list)
{
#if defined RUNTIME_DEBUG && defined USE_POSIX
	printf("Lock port list\n");
#endif
  return pthread_mutex_lock(port_list->rez);
}

StatusType unlock_port_list(port_list_t * port_list)
{
#if defined RUNTIME_DEBUG && defined USE_POSIX
	printf("Unlock port list\n");
#endif
  return pthread_mutex_unlock(port_list->rez);
}

StatusType wait_port_list(port_list_t * port_list)
{
  int ret = pthread_cond_wait(port_list->event,
                              port_list->rez);
  return ret;
}

